/* Copyright (c) 2008, XenSource Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of XenSource Inc. nor the names of its contributors
 *       may be used to endorse or promote products derived from this software
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#ifndef __BLKTAP2_VHD_UUID_H__
#define __BLKTAP2_VHD_UUID_H__

#if defined(__linux__)

#include <uuid/uuid.h>
typedef struct {
    uuid_t uuid;
} vhd_uuid_t;

#elif defined(__NetBSD__)

#include <uuid.h>

typedef uuid_t vhd_uuid_t;

#else

#error "Please update vhd-uuid.h for your OS"

#endif

int vhd_uuid_is_nil(vhd_uuid_t *uuid);

void vhd_uuid_generate(vhd_uuid_t *uuid);

void vhd_uuid_to_string(vhd_uuid_t *uuid, char *out, size_t size);

void vhd_uuid_from_string(vhd_uuid_t *uuid, const char *in);

void vhd_uuid_copy(vhd_uuid_t *dst, vhd_uuid_t *src);

void vhd_uuid_clear(vhd_uuid_t *uuid);

int vhd_uuid_compare(vhd_uuid_t *uuid1, vhd_uuid_t *uuid2);

#endif /* __BLKTAP2_VHD_UUID_H__ */
