#ifndef __ASM_ARM_BYTEORDER_H__
#define __ASM_ARM_BYTEORDER_H__

#define __BYTEORDER_HAS_U64__

#include <xen/byteorder/little_endian.h>

#endif /* __ASM_ARM_BYTEORDER_H__ */
/*
 * Local variables:
 * mode: C
 * c-file-style: "BSD"
 * c-basic-offset: 4
 * indent-tabs-mode: nil
 * End:
 */
