#ifndef __ASM_PERFC_H__
#define __ASM_PERFC_H__

static inline void arch_perfc_reset(void)
{
}

static inline void arch_perfc_gather(void)
{
}

#endif

/*
 * Local variables:
 * mode: C
 * c-file-style: "BSD"
 * c-basic-offset: 4
 * indent-tabs-mode: nil
 * End:
 */
