/******************************************************************************
 * nmi.h
 *
 * Register and unregister NMI callbacks.
 *
 * Copyright (c) 2006, Ian Campbell <ian.campbell@xensource.com>
 */

#ifndef __XEN_NMI_H__
#define __XEN_NMI_H__

#include <asm/nmi.h>

#endif /* __XEN_NMI_H__ */
